proc UniformErr {} {
    set err [new ErrorModel/CL]
    $err unit packet
    return $err
}


Node/MobileNode set default_error_ .00
Node/MobileNode set max_channels_ 12
Node/MobileNode set switch_time_ .001
ErrorModel/CL set Bandwidth 2000000
ErrorModel/CL set BitRate 1000000
#power and noise values taken from "Efficient Geographic Routing in Multihop Wireless Networks", S. Lee, B. Bhatacharjee, S. Banerjee
Phy/WirelessPhy set ambient_noise_ .0000000000012
Phy/WirelessPhy set Pt_ 0.020
Phy/WirelessPhy set CSThresh_ 10
Mac/802_11 set dataRate_ 1Mb
Mac/802_11 set PreambleLength_ 72

#clear_rt_on_switch indicates if the routing tables should be cleared when an interface changes channels. 1 = clear routing tables, 0 = leave routing tables
Node/MobileNode set clear_rt_on_switch_         1;



set val(chan)           Channel/WirelessChannel    ;# Channel Type
set val(prop)           Propagation/TwoRayGround   ;# radio-propagation model
set val(netif)          Phy/WirelessPhy            ;# network interface type
set val(mac)            Mac/802_11                 ;# MAC type
set val(ifq)            Queue/DropTail/PriQueue    ;# interface queue type
set val(ll)             LL                         ;# link layer type
set val(ant)            Antenna/OmniAntenna        ;# antenna model
set val(ifqlen)         50                         ;# max packet in ifq
set val(nn)             6                          ;# number of mobilenodes
set val(rp)             AODV                  ;# routing protocol
set val(x)		800
set val(y)		800
set val(interfaces)     2                    ;#number of interfaces each node has
set val(length)         800                  ;#length of simulation


# Initialize Global Variables
set ns_		[new Simulator]
set tracefd     [open ./2w_mitf.tr w]
#$ns_ use-newtrace
$ns_ trace-all $tracefd


# set up topography object
set topo       [new Topography]

$topo load_flatgrid $val(x) $val(y)

# Create God
create-god [expr $val(nn) * ($val(interfaces) + 1)]

# Create channels
#notice how channels are created ahead of time
for {set i 0} {$i < 15 } {incr i} {
    set chan($i) [new $val(chan)] 
}


#=====================================	
#must set Uniform error to get gray model
set inerrProc_ "UniformErr"
set outerrProc_ ""
set FECProc_ ""
set propInstance_ [new $val(prop)]
set port [Node set rtagent_port_]


# configure node, please note the change below.
$ns_ node-config -adhocRouting $val(rp) \
		-llType $val(ll) \
		-macType $val(mac) \
		-ifqType $val(ifq) \
		-ifqLen $val(ifqlen) \
		-antType $val(ant) \
		-propType $val(prop) \
		-phyType $val(netif) \
		-topoInstance $topo \
		-agentTrace ON \
		-routerTrace OFF \
		-macTrace OFF \
		-movementTrace OFF \
		-channel $chan(0) \
                -IncomingErrProc UniformErr


for {set i 0} {$i < $val(nn) } {incr i} {
    set node_($i) [$ns_ node]  # create the node
    set agent [$node_($i) set ragent_]  # get the routing agent
    for {set j 0} {$j < $val(interfaces) } {incr j} {
	# add an interface on the next channel
	$node_($i) add-interface $chan([expr {$j + 1}]) $propInstance_ $val(ll) $val(mac)	\
	    $val(ifq) $val(ifqlen) $val(netif) $val(ant) \
	    $topo $inerrProc_ $outerrProc_ $FECProc_
	#this is stupid and redundent, but I'm not going to spend the time rewritting it
	$node_($i) adjust-target-rtagent $agent $port
    }
    $node_($i) random-motion 0
}



set tcp1 [new Agent/TCP]
set sink1 [new Agent/TCPSink]
$node_(1) attach $tcp1 30 1
$node_(3) attach $sink1 30 1
$ns_ connect $tcp1 $sink1
set ftp1 [new Application/FTP]
$ftp1 attach-agent $tcp1

#command to switch channel of a node 
# note this is also exposed through the C interface
# while switching nodes will queue any packets to send and will drop any incomming pacets
#node switch-channel (interface to change) (destination channel)
$ns_ at 8.5 "$node_(4) switch-channel 1 11"
$ns_ at 8.5 "$node_(5) switch-channel 1 11"
$ns_ at 8.5 "$node_(4) switch-channel 2 0"
$ns_ at 8.5 "$node_(5) switch-channel 2 0"
$ns_ at 30.1 "$ftp1 produce 700"
$ns_ at 10.0 "puts Sending"

$ns_ at $val(length) "stop"
$ns_ at [expr {$val(length) +.0001}] "puts \"NS EXITING...\" ; $ns_ halt"

proc stop {} {
    global ns_ tracefd
    $ns_ flush-trace
    close $tracefd
}

puts "Starting Simulation..."
$ns_ run
